#!/usr/bin/python
# All Rights Reserved. Copyright (c) 2018, Hitachi, Ltd.       #
# Licensed Material of Hitachi, Ltd.                           #
# Reproduction, use, modification or disclosure otherwise than #
# permitted in the License Agreement is strictly prohibited.   #

from __future__ import with_statement
import os
import errno
import base64
import hashlib
import xml.etree.ElementTree as ET
from xml.parsers.expat import ExpatError

import HttpClient
try:
    from CommonLogger import log_debug, log_info, log_warning, log_error
except ImportError:
    # To make this module handy and portable
    def log_debug(msg): pass
    def log_info(msg): pass
    def log_warning(msg): print msg
    def log_error(msg): print msg

class BaseError(Exception):
    def __init__(self, reason, code=None):
        self.reason = reason.strip()
        self.code = code

    def __str__(self):
        if self.code is None:
            return "%s" % self.reason
        else:
            return "%s (code=%s)" % (self.reason, self.code)

class FormatError(BaseError):
    pass

class ConfigurationError(BaseError):
    pass

class ConnectionError(BaseError):
    pass

class ParameterError(BaseError):
    pass

class HCPInfo(object):
    def __init__(self):
        self.domain = None
        self.tenant = None
        self.user64 = None
        self.passMD5 = None
        self.rephost = None
        self.exhost = None
        self.repexhost = None
        self.ssl = None
        self.proxy = None   # Dictonary

    def load(self):
        # Load HCP configuration file on the node, assuming I am on an HDI
        hcp_config_file = "/enas/hsmarc/conf/dataaccess.conf"
        ssl_config_file = "/enas/hsmarc/conf/sslctrl.conf"
        proxy_config_file = "/enas/hsmarc/conf/hcp_proxy.conf"

        # HCP Connection information
        try:
            tree = ET.parse(hcp_config_file)
        except EnvironmentError, e:
            if e.errno == errno.ENOENT:
                log_error("HCP configuration file does not exist on this node. file=%s" % hcp_config_file)
                raise FormatError("%s does not exist." % hcp_config_file)
            else:
                log_error("Unable to open/read HCP configuration file. file=%s, error=%s" % (hcp_config_file, str(e)))
                raise
        except ExpatError, e:
            log_error("Invalid XML syntax in HCP configuration file. file=%s, code=%s, line=%s, offset=%s, error=%s" % (hcp_config_file, e.code, e.lineno, e.offset, str(e)))
            raise FormatError("Invalid XML syntax in HCP configuration file.")

        root = tree.getroot()
        try:
            hcpconfig = root.find('HCAPConfig')
            self.domain = hcpconfig.attrib['hostName']
            self.tenant = hcpconfig.attrib['tenantName']

            account = root.find('Account')
            self.user64 = account.attrib['username']
            self.passMD5 = account.attrib['password']

            repinfo = root.find('ReplicaConfig')
            if len(repinfo.attrib['hostName']) != 0:
                self.rephost = repinfo.attrib['hostName']

            exinfo = root.find('ExHost')
            if len(exinfo.attrib['hostName']) != 0:
                self.exhost = exinfo.attrib['hostName']
            if len(exinfo.attrib['replicaHostName']) != 0:
                self.repexhost = exinfo.attrib['replicaHostName']
        except (AttributeError, KeyError), e:
            log_error("Expected information is missing in HCP configuration file. file=%s, error=%s" % (hcp_config_file, str(e)))
            raise FormatError("Expected information is missing in HCP configuration file.")

        # SSL setting
        self.ssl = os.path.exists(ssl_config_file)

        # Proxy setting
        try:
            tree = ET.parse(proxy_config_file)
        except EnvironmentError, e:
            if e.errno == errno.ENOENT:
                log_error("HCP proxy configuration file does not exist on this node. file=%s" % proxy_config_file)
                raise FormatError("%s does not exist." % proxy_config_file)
            else:
                log_error("Unable to open/read HCP proxy configuration file. file=%s, error=%s" % (proxy_config_file, str(e)))
                raise
        except ExpatError, e:
            log_error("Invalid XML syntax in HCP proxy configuration file. file=%s, code=%s, line=%s, offset=%s, error=%s" % (proxy_config_file, e.code, e.lineno, e.offset, str(e)))
            raise FormatError("Invalid XML syntax in HCP proxy configuration file.")

        HCPProxy = tree.getroot()
        try:
            if len(HCPProxy.attrib['host']) != 0:
                self.proxy = {'host': HCPProxy.attrib['host'],
                              'port': int(HCPProxy.attrib['port']),
                              'user': HCPProxy.attrib['username'] if len(HCPProxy.attrib['username']) != 0 else None,
                              'pass': HCPProxy.attrib['password'] if len(HCPProxy.attrib['password']) != 0 else None
                             }
        except (AttributeError, KeyError), e:
            log_error("Expected information is missing in HCP proxy configuration file. file=%s, error=%s" % (proxy_config_file, str(e)))
            raise FormatError("Expected information is missing in HCP proxy configuration file.")

        if self.proxy is not None:
            proxy_secret = self.proxy.copy()
            proxy_secret['pass'] = "***"
        else:
            proxy_secret = None
        log_info("HCP info: domain=%s, tenant=%s, exhost=%s, ssl=%s, proxy=%s" % (self.domain, self.tenant, self.exhost, str(self.ssl), proxy_secret))

    def set_domain(self, domain):
        self.domain = domain

    def set_tenant(self, tenant):
        self.tenant = tenant

    def set_account(self, user, password=None, passwordMD5=None):
        self.user64 = base64.b64encode(user)
        if password is not None:
            m = hashlib.md5()
            m.update(password)
            self.passMD5 = m.hexdigest()
        elif passwordMD5 is not None:
            self.passMD5 = passwordMD5

    def set_exhost(self, exhost, repexhost=None):
        self.exhost = exhost
        self.repexhost = repexhost

    def set_replica(self, rephost):
        self.rephost = rephost

    def set_proxy(self, host=None, port=None, user=None, password=None, proxy=None):
        if proxy is not None:
            self.proxy = proxy
        else:
            self.proxy = {'host': host,
                          'port': int(port),
                          'user': user,
                          'pass': base64.b64encode(password) if password is not None else None
                         }

    def enable_ssl(self):
        self.ssl = True

    def disable_ssl(self):
        self.ssl = False

class ContentFormat(object):
    # python2.5 doesn't have Enum...
    XML = 0
    JSON = 1

class HCPClient(object):
    def __init__(self, hcp_info, namespace=None, port=None):
        self.hcp_info = hcp_info
        self.namespace = namespace
        self.port = port
        self.format_send = ContentFormat.XML
        self.format_recv = ContentFormat.XML
        self.prettyprint = False
        self.verbose = False
        self.version = None
        self.version_list = False
        self.deleted = False
        self.params = None
        self.retry_count = 0
        self.retry_interval = 0
        self.timeout = 0
        self.use_content_header = False

    def set_connection(self, count, interval, timeout):
        self.retry_count = count
        self.retry_interval = interval
        self.timeout = timeout

    def enable_prettyprint(self):
        self.prettyprint = True

    def disable_prettyprint(self):
        self.prettyprint = False

    def enable_verbose(self):
        self.verbose = True

    def disable_verbose(self):
        self.verbose = False

    def enable_content_header(self):
        self.use_content_header = True

    def disable_content_header(self):
        self.use_content_header = False

    def set_send_format(self, fmt):
        if fmt != ContentFormat.XML and fmt != ContentFormat.JSON:
            log_error("Invalid value is specified for HCPClient.set_send_format(). fmt=%s" % fmt)
            raise ParameterError("Invalid Content-type: %s" % fmt)
        self.format_send = fmt

    def set_recv_format(self, fmt):
        if fmt != ContentFormat.XML and fmt != ContentFormat.JSON:
            log_error("Invalid value is specified for HCPClient.set_recv_format(). fmt=%s" % fmt)
            raise ParameterError("Invalid Accept: %s" % fmt)
        self.format_recv = fmt

    def set_version(self, version):
        if version is None:
            self.version = None
        else:
            self.version = {'start': int(version), 'end': None, 'istime': False, 'isrange': False}
        self.version_list = False

    def set_version_with_time(self, time):
        self.version = {'start': int(time), 'end': None, 'istime': True, 'isrange': False}
        self.version_list = False

    def set_version_range(self, start, end=None):
        if end is None:
            self.version = {'start': int(start), 'end': None, 'istime': False, 'isrange': True}
        else:
            self.version = {'start': int(start), 'end': int(end), 'istime': False, 'isrange': True}
        self.version_list = False

    def set_version_range_with_time(self, start, end=None):
        if end is None:
            self.version = {'start': int(start), 'end': None, 'istime': True, 'isrange': True}
        else:
            self.version = {'start': int(start), 'end': int(end), 'istime': True, 'isrange': True}
        self.version_list = False

    def enable_version_list(self):
        self.version_list = True
        self.version = None

    def set_deleted(self, deleted):
        self.deleted = deleted

    def set_params(self, params):
        self.params = params

    def _exec_method(self, method, path, from_file=None, data=None, to_file=None):
        if method == HttpClient.HTTPMethod.PUT:
            if from_file is None and data is None:
                raise ParameterError("No data source for PUT request.")

        baseurl = self._make_base_url()
        try:
            client = self._setup_Httpclient(baseurl)
            if method == HttpClient.HTTPMethod.HEAD:
                ret = client.head(path)
            elif method == HttpClient.HTTPMethod.GET:
                ret = client.get(path, to_file)
            elif method == HttpClient.HTTPMethod.PUT:
                ret = client.put(path, from_file=from_file, data=data)
            elif method == HttpClient.HTTPMethod.POST:
                ret = client.post(path, from_file=from_file, data=data, to_file=to_file)
            elif method == HttpClient.HTTPMethod.DELETE:
                ret = client.delete(path)
        except HttpClient.ConnectionError, e:
            log_error("Failed to connect the HCP. host=%s, path=%s, method=%s, code=%s, reason=%s." % (baseurl, path, HttpClient.HTTPMethod.tostr(method), e.code, e.reason))
            raise ConnectionError(e.reason, e.code)

        return ret


    def head(self, path):
        return self._exec_method(HttpClient.HTTPMethod.HEAD, path)

    def get(self, path, to_file=None):
        return self._exec_method(HttpClient.HTTPMethod.GET, path, to_file=to_file)

    def put(self, path, from_file=None, data=None):
        return self._exec_method(HttpClient.HTTPMethod.PUT, path, from_file=from_file, data=data)

    def post(self, path, from_file=None, data=None, to_file=None):
        return self._exec_method(HttpClient.HTTPMethod.POST, path, from_file=from_file, data=data, to_file=to_file)

    def delete(self, path):
        return self._exec_method(HttpClient.HTTPMethod.DELETE, path)

    def exec_MQE(self, query, to_file=None):
        baseurl = self._make_base_url()
        path = "/query"
        try:
            client = self._setup_Httpclient(baseurl)
            ret = client.post(path, data=query, to_file=to_file)
        except HttpClient.ConnectionError, e:
            log_error("Failed to connect the HCP. host=%s, path=%s, method=%s, code=%s, reason=%s." % (baseurl, path, 'POST', e.code, e.reason))
            raise ConnectionError(e.reason, e.code)
        return ret

    def _set_auth_header(self, client):
        auth_token = "%s:%s" % (self.hcp_info.user64, self.hcp_info.passMD5)
        client.add_header(('Authorization', ("HCP %s" % auth_token)))

    def _set_content_header(self, client):
        if self.format_send == ContentFormat.XML:
            client.add_header(('Content-Type', "application/xml"))
        elif self.format_send == ContentFormat.JSON:
            client.add_header(('Content-Type', "application/json"))

        if self.format_recv == ContentFormat.XML:
            client.add_header(('Accept', "application/xml"))
        elif self.format_recv == ContentFormat.JSON:
            client.add_header(('Accept', "application/json"))

    def _make_base_url(self):
        if self.hcp_info.ssl is True:
            protocol = "https"
        else:
            if self.namespace is None:
                # Access to a tenant means MAPI or MQE access - always use https for these cases
                protocol = "https"
            else:
                protocol = "http"

        if self.namespace is None:
            host = "%s.%s" % (self.hcp_info.tenant, self.hcp_info.domain)
        else:
            host = "%s.%s.%s" % (self.namespace, self.hcp_info.tenant, self.hcp_info.domain)

        baseurl = {}
        baseurl['protocol'] = protocol
        baseurl['host'] = host
        baseurl['port'] = self.port

        return baseurl

    def _setup_Httpclient(self, baseurl):
        # HttpClient
        if self.hcp_info.exhost is not None:
            # Use external host
            target_host = self.hcp_info.exhost
        else:
            # Do not use external host
            target_host = baseurl['host']

        host = "%s://%s" % (baseurl['protocol'], target_host)
        port = ":%s" % baseurl['port'] if baseurl['port'] is not None else ""
        host += port
        client = HttpClient.HttpClient(host)

        if self.hcp_info.exhost is not None:
            client.add_header(('Host', baseurl['host']))

        self._set_auth_header(client)
        if self.use_content_header:
            self._set_content_header(client)

        # Proxy
        if self.hcp_info.proxy is not None:
            if self.hcp_info.proxy['user'] is None:
                client.set_proxy(self.hcp_info.proxy['host'], self.hcp_info.proxy['port'])
            else:
                client.set_proxy(self.hcp_info.proxy['host'], self.hcp_info.proxy['port'], self.hcp_info.proxy['user'], self.hcp_info.proxy['pass'])

        # Set version range
        if not self.version is None:
            if self.version['istime'] is False and self.version['isrange'] is False:
                # A version ID is specified
                client.add_param({'version' : self.version['start']})
            elif self.version['istime'] is False and self.version['isrange'] is True:
                # Range of version ID is specified
                if self.version['end'] is None:
                    vrange = "%s-" % self.version['start']
                else:
                    vrange = "%s-%s" % (self.version['start'], self.version['end'])
                client.add_param({'version' : vrange})
            elif self.version['istime'] is True and self.version['isrange'] is False:
                # A time is specified
                vtime = "@%s" % self.version['start']
                client.add_param({'version' : vtime})
            elif self.version['istime'] is True and self.version['isrange'] is True:
                # Range of time is specified
                if self.version['end'] is None:
                    vrange = "@%s-" % self.version['start']
                else:
                    vrange = "@%s-@%s" % (self.version['start'], self.version['end'])
                client.add_param({'version' : vrange})
        elif self.version_list:
            client.add_param({'version': 'list'})

        if self.deleted is True:
            client.add_param({'deleted': 'true'})

        if self.prettyprint is True:
            client.add_param({'prettyprint' : None})
        if self.verbose is True:
            client.add_param({'verbose' : 'true'})
        if self.params is not None:
            for key in self.params.keys():
                client.add_param({key: self.params[key]})
        client.set_retry(self.retry_count, self.retry_interval)
        client.set_timeout(self.timeout)

        return client
